from __future__ import annotations

import ast
from abc import abstractmethod, ABC
from typing import Tuple, List, Type, Iterator, TYPE_CHECKING

if TYPE_CHECKING:
    from .violation import Violation, ViolationFactory


class Visitor(ast.NodeVisitor):
    def __init__(self, plugin: Plugin):
        super().__init__()
        self.plugin = plugin
        self.violate = plugin.violate


class Plugin(ABC):
    @property
    @abstractmethod
    def name(self) -> str:
        ...

    @property
    @abstractmethod
    def visitors(self) -> List[Type[Visitor]]:
        ...

    def __init__(self, tree: ast.AST):
        self._tree = tree
        self.violations: List[Violation] = []

    def violate(self, violation_type: ViolationFactory, node: ast.AST, **kwargs) -> None:
        violation = violation_type(node.lineno, node.col_offset, kwargs)
        self.violations.append(violation)

    def _run(self) -> List[Violation]:
        for visitor_type in self.visitors:
            visitor = visitor_type(self)
            visitor.visit(self._tree)

        return self.violations

    def run(self) -> Iterator[Tuple[int, int, str, type]]:
        yield from (violation.as_tuple(self.__class__) for violation in self.violations)



import ast
from textwrap import dedent
from typing import Type, Optional

from .plugin import Plugin
from .violation import Violation


def _violation_from_src(plugin_type: Type[Plugin], src: str):
    tree = ast.parse(dedent(src))
    plugin = plugin_type(tree)
    violations = plugin._run()
    if len(violations) == 0:
        return None

    assert len(violations) == 1
    return violations[0]


def assert_violates(plugin_type: Type[Plugin], expected_violation: Optional[Violation], src: str):
    violation = _violation_from_src(plugin_type, src)
    assert violation == expected_violation, f'Expected {expected_violation}, got {violation}'


def assert_not_violates(plugin_type: Type[Plugin], src: str):
    assert_violates(plugin_type, None, src)



from __future__ import annotations

from dataclasses import dataclass, field
from functools import partial
from typing import Type, Dict, Any, Callable, TYPE_CHECKING, List

if TYPE_CHECKING:
    from .plugin import Plugin


@dataclass
class Violation:
    code: int
    message: str
    line: int
    col: int
    args: List[Any] = field(default_factory=list)
    kwargs: Dict[str, Any] = field(default_factory=dict)

    @property
    def full_message(self):
        return f'{str(self.code).zfill(3)} {self.formatted_message}'

    @property
    def formatted_message(self):
        return self.message.format(*self.args, **self.kwargs)

    def as_tuple(self, type_: Type[Plugin]):
        return self.line, self.col, self.full_message, type_


create_violation = partial(partial, Violation)
ViolationFactory = Callable[[int, int, Dict[str, Any]], Violation]
